/* BTB.H
 * Private/Shared branch target buffer for multithreaded processor
 * Parameters:
 *   Size
 *   PC-tagged/Cacheline-tagged/Untagged
 *   PC-returning/Cacheline-returning
 */

#ifndef AINT_BTB_H
#define AINT_BTB_H

#include <iostream.h>
#include <sys/types.h>

typedef enum BTB_TAG_e
  {BTB_TAG_invalid=0, BTB_TAG_PC, BTB_TAG_CacheIndex, BTB_TAG_None}
  BTB_TAG_type;

class BTB {
public:
  int Size;
  BTB_TAG_type TagType;
  int CacheIndexBits;
  int DataIsPCnotCLN;

  ulong CacheIndexMask;

  ulong *Tags;
  ulong *Data;
  uchar *Valid;

  ulong Refs;
  ulong Misses;

public:
  BTB(int size, BTB_TAG_type tagtype, int dataispc, int cacheindexbits)
  {
    Size=size;
    TagType=tagtype;
    DataIsPCnotCLN = dataispc;

    CacheIndexBits = cacheindexbits;
    CacheIndexMask = (1<<CacheIndexBits)-1;

    Refs=0;
    Misses=0;

    if ( (TagType < 1) || (TagType > BTB_TAG_None) ) {
      cerr << "Invalid Tag Type" << endl;
      exit(1);
    }

    // Allocate tag array if tagged, of appropriate size
    if (TagType!=BTB_TAG_None) {
      Tags = new ulong[Size];
      Valid = new uchar[Size];
    }

    // Allocate data array
    Data = new ulong[Size];

  }

  ulong Lookup(ulong PC)
  {
    int index = (PC>>2) & (Size-1);

    Refs++;

    if (TagType != BTB_TAG_None) {
      return Data[index];
    } else {
      // Get correct tag value
      ulong tag = (TagType==BTB_TAG_PC) ? PC : ((PC>>2)&CacheIndexMask);
      if ((Tags[index] == tag) && Valid[index]) {
	return (ulong) Data[index];
      } else {
	return 0x0;
      }
    }
  }

  void Update(ulong PC, ulong Target, int Hit)
  {
    int index = (PC>>2) & (Size-1);
    int PC_CLN = ( (PC>>2) & CacheIndexMask);
    int Target_CLN = ( (Target>>2) & CacheIndexMask);
    ulong data = DataIsPCnotCLN ? Target : Target_CLN;

    if (!Hit) Misses++;

    Data[index] = data;

    if (TagType == BTB_TAG_PC) {
      Tags[index] = PC;
      Valid[index] = 1;
    } else if (TagType == BTB_TAG_CacheIndex) {
      Tags[index] = PC_CLN;
      Valid[index] = 1;
    }
  }

};

#endif
