

// branch.C: a simple non-pipelined program-order processor with a branch
// predictor

/* This simulates an N-entry 2-bit branch predictor indexed by PC and
 * reports the number of hits, misses and mispredict rate
 */

#include <iostream.h>
#include <stdlib.h>

#include <aint.h>
// The catch-all include file that declares all relevant AINT types, consts,
// variables, and objects

aint_addr_t PC;    // address to fetch from
path_ptr ThePath;  // Main path in the main thread in the target binary

const int DefaultMaxL = 16;
ulong N;

// Two bits per line
uchar *Counters;
ulong Insts, Branches, Mispredicts;
ulong BHR;

int Taken[] = {1, 2, 3, 3};
int NotTaken[] = {0, 0, 1, 2};

const int Kilos = (1 << 10);
const int Megs = (1 << 20);
const int DefaultMaxInsts = (1*Megs);

int MaxInsts;
int MaxL;

// Aint calls this function before opening any object files - use this
// to initialize back-end variables
int
sim_init(int argc, char **argv, char **envp)
{
  /* AINT uses the getopt() function to parse the argument string, and
   * expects optind to have the correct value (first object file name)
   * upon return from this function. We must therefore use getopt() to
   * process arguments meant for us (delimited by "--").
   */

  extern int optind;    // AINT uses the getopt() function to keep track of args
  int i;
  char c;
  // For this example, we use no arguments...

  Insts = 0;

  MaxInsts = DefaultMaxInsts;
  MaxL = DefaultMaxL;

  while( (c = getopt(argc, argv, "n:l:")) != -1) {
    switch(c) {
    case 'n':
      MaxInsts = strtol(optarg, NULL, 0);
      break;
    case 'l':
      MaxL = strtol(optarg, NULL, 0);
      if (MaxL < 0) MaxL = -1;
      break;
    default:
      sim_usage(argc, argv);
    }
  }

  cerr << "Max Insts = " << MaxInsts << "; Size of index into counter table = " << MaxL << "bits." << endl;

  N = 1<<MaxL;
  Counters = new uchar[N];

  // Initialize the predictor lines - all counters to Weakly not-taken (01)
  for (i=0; i<N; i++) {
    Counters[i] = 0x1;
  }
  BHR = 0x0;
  Branches = 0;
  Mispredicts = 0;

  // Return optind so that AINT finds the name of the target binary...
  return optind;
}


// Aint calls this function if the command line has an error
void
sim_usage(int argc, char **argv)
{
  cerr << endl << "Usage: <simname> [AINT options] executable" << endl;
  cerr << "This simulator takes no arguments, and accepts one executable" << endl;
}


// AINT calls this function when it creates the main thread in the process
// corresponding to the target binary
void
sim_thread_begin(int tid, aint_addr_t pc)
{
  thread_ptr pthread = Threads[tid];    // Get thread pointer from AINT's table
  ThePath = AINT_create_path(tid, pc);    // Ask AINT to create main path

  informative("sim_thread_begin: began thread/path" << pthread->tid
      << " " << (void *) pthread->path_tree->pathid << endl);

  PC = pc;    // Starting address for thread
}


// The main execution loop - fetch, execute, commit each inst one at a time
void
sim_execute()
{
  inflight_inst_ptr ifi;

  informative("sim_execute: commencing execution at " << (void *) PC << endl);

  while(PC && (Insts < MaxInsts)) {    // we get a zero PC in *ifi when program executes _exit()

    ulong Tag;
    int PredBits;

    ifi = AINT_fetch(ThePath, PC);    // fetch next instruction
    informative("Fetch: " << (void *) PC << endl);
    Insts++;

    if (ifi->iflags & E_CONDBR) {
	Tag = (PC>>2) & (N-1);
	Tag ^= BHR;
	PredBits = Counters[Tag];
    }

    AINT_issue(ifi);    // ask AINT to issue the fetched instruction
    informative("Issue: " << (void *) PC << endl);

    // Check prediction and update if necessary
    if (ifi->iflags & E_CONDBR) {

      if ((PredBits >> 1) != (ifi->nextPC != PC + 4)) {
	// mispredict! 
        Mispredicts++;
      } else {
        // predicted correctly!
      }

      BHR = (BHR << 1) & (N-1);

      if (ifi->nextPC == PC+4) {
        PredBits = NotTaken[PredBits];
      } else {
        PredBits = Taken[PredBits];
	BHR = BHR | 1;
      }
      Counters[Tag] = PredBits;

      Branches++;
    }
    if (ifi->iflags & E_READ) {    // this is a load inst

      AINT_do_read(ifi);

    } else if (ifi->iflags & E_WRITE) {    // this is a store inst

      AINT_do_write(ifi);

    }

    PC = ifi->nextPC;    // AINT updates ifi->nextPC upon executing inst

    AINT_commit(ifi);    // Done
  }

}  // sim_execute()



void
sim_done()
{
  cout << "Insts =" << Insts << " Branches=" << Branches << "; Misses=" << Mispredicts << " Miss Rate: "
       << ((float)Mispredicts/(float)Branches) << endl;

}

