/* process.h
 *
 * The process is at the top of the context hierarchy
 * supported in AINT. The process contains the structures for address
 * space implementation. Within the process are threads. A thread is
 * the logical control stream seen by the programmer.
 * Within threads are paths - streams of execution created at diffident
 * branches
 */

#ifndef AINT_PROCESS_H
#define AINT_PROCESS_H

#include <sys/mman.h>

#include "tlb.h"


/* corresponds to an object file section or to a region allocated by shmat or mmap */
typedef class segment {
public:
  aint_addr_t start;     /* in object space */
  aint_addr_t end;       /* inclusive: end = (start + size-1) */
  ulong size;
  int prot;        /* use mmap() PROT_xxxx flags */
  int flags;       /* use mmap() MAP_xxxx flags */

  int contains(aint_addr_t _addr);

} segment_t;



/* Per-process shared-region descriptors */
struct Shm_ds {
  struct Shm_ds *next;
  long shmid;    /* Index into the global (ipc) shmem struct */
  long size;
  aint_addr_t addr;    /* Base address to which mapped */
};


#define SETSIG(sigvec, signum)    (sigvec |= (1 << ((signum)-1)))
#define CLRSIG(sigvec, signum)    (sigvec &= ~(1<<((signum)-1)))
#define ISSETSIG(sigvec, signum) (sigvec & (1 <<((signum)-1)))

typedef void (*sig_handler_t)(int, ...);

struct my_sigvec {
  sig_handler_t sv_handler;
  int sv_mask;
  int sv_flags;
};


class process {
public:

  process(const char * _Objname); // For a brand new process

  process(process_ptr parent, thread_ptr parent_thread);    // In fork()

  int pid;

  // pointers used to establish relations between processes
  process_ptr parent;
  process_ptr youngest;
  process_ptr sibling;  // next older sibling

  TLB *tlb;
  int num_pages;        /* Number of page-table entries allocated - hack for speeding up fork */
  int num_private;        /* Number of actual pages allocated as private */

  struct Shm_ds *Shmem_regions;    /* List of shared regions attached by process */
  ulong Unsp_Shmat_Current;    /* Address to use in shmat called with NULL addr */

  // Threads belonging to this process:
  thread_ptr threads;    /* list of threads running in this process. Main is first */
  thread_ptr youngest_thread;  /* Last in the list of threads - for fast append */
  int thread_count;

  state_t runstate;
  struct my_sigvec *sigv;    /* for signal handlers */


  segment_t *segments; /* memory segments used by this process */
  int n_segments;

  // Information from the object file:
  char *Objname;
  int n_instructions;

  aint_addr_t entry_point;   /* entry point for this process (virtual address) */
  aint_addr_t initial_gp;    /* GP value for this process upon entry */
  aint_addr_t stack_end;
  aint_addr_t Data_end;
  aint_addr_t brk;
  aint_addr_t sim_user_addr;    /* address of the sim_user function in the executable */

  int use_dynamic_loader;
  int fd[MAX_FDNUM];    /* file descriptors */

  int pthread_keys_in_use;    /* # of TSD keys created */

  // member functions for dealing with object files
  // void load_object_file();
  void create_addr_space();
  int copy_argv(int, int, char**, char**);
  void decode_text_region(ulong start, ulong size);

  segment_t *add_segment(aint_addr_t start, ulong size, int prot, int flags);
  segment_t *find_segment(aint_addr_t addr);

  inline aint_addr_t allocate_thread_stack(int tid);

  inline void * addr2phys(aint_addr_t _vaddr, long *paccess);
  void * PMAP(aint_addr_t _vaddr) {return addr2phys(_vaddr, NULL); }

  icode_ptr addr2iphys(aint_addr_t _vaddr, long *paccess);
 
  void memcpy_to_object_space(aint_addr_t object_ptr, void *aint_ptr, size_t size);
  void memcpy_from_object_space(void *aint_ptr, const aint_addr_t object_ptr, size_t size);
  size_t strlen_from_object_space(aint_addr_t object_ptr);
  char * strdup_from_object_space(aint_addr_t object_ptr);
  void strcpy_from_object_space(char *aint_ptr, aint_addr_t object_ptr);
  void strcpy_to_object_space(aint_addr_t object_ptr, const char *aint_ptr);

};


// definition of constants pertaining to address spaces and object files


inline void *
process::addr2phys(aint_addr_t addr, long *paccess)
{
  /* Refer to the Local TB First */
  tb_entry *pg;

  pg = tlb->Translate(addr, paccess);

  if (!pg) return NULL;

  return (void *) ((ulong)pg->page + TB_OFFSET(addr));
}


inline aint_addr_t
process::allocate_thread_stack(int tid)
{
  /* extend stack to create space for thread */
  ulong new_stack_end = Data_end - (tid*Stack_size);
  add_segment(new_stack_end - Stack_size,
	      Stack_size,
	      PROT_READ|PROT_WRITE,
              MAP_PRIVATE);
  return new_stack_end;
}




#endif
